# ================================================================
# Name:      KeyGenerator.py
# Purpose:   Password generator.
# Author:    Joe Linoff
# Created:   2012-02-03
# Id:        $Id: KeyGenerator.py,v 1.2 2012/02/04 02:23:48 jlinoff Exp jlinoff $
# Copyright: (c) 2012 Joe Linoff
# License:
#     SecEd is free software; you can redistribute 
#     it and/or modify it under the terms of the GNU General Public License as 
#     published by the Free Software Foundation; either version 2 of the License, 
#     or (at your option) any later version.
#            
#     SecEd is distributed in the hope that it will be useful, 
#     but WITHOUT ANY WARRANTY; without even the implied warranty of 
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#     See the GNU General Public License for more details. You should have 
#     received a copy of the GNU General Public License along with SecEd; 
#     if not, write to the Free Software Foundation, Inc., 59 Temple Place, 
#     Suite 330, Boston, MA  02111-1307  USA
# ================================================================
import re
import random

class CharSet:
    """
    Manages character sets. A character set is specified by a comma
    separated tuple of modifiers and a single expression in a
    string. Here is an example that will explained later.

        # accept digits, lower case letters, upper case letters,
        # dollar sign and period. At least 3 must appear.
        charset #1: '[0-9a-bA-Z$.]',min-matches=3

        # exclude lower case L and upper case o
        charset #2: '[lO]',min-matches=0

    If there is a syntax error in the string an exception is thrown.
    You can catch it as follows:

        arg = '[0-9a-bA-Z$.],min-matches=3'
        try:
            charset = CharSet(arg,True).Init()
            charsets.append(charset)
        except Exception as (strerr):
            print 'ERROR: charset compilation failed: %s' % (strerr)

    The expression consists of printable characters in the ASCII range
    of 32 (' ') to 126 ('~') enclosed in brackets. A range can be
    specified using a dash ('-'). A dash is specified by inserting it
    at the front. Internal brackets do not have to be escaped.

    Here are a couple examples of expressions:

        Expression         Description
        =================  ======================================
        [0123456789]       All digits.
        [0-9]              All digits.
        [a-z0-9]           All lower case letters and all digits.
        [-[]]              A dash and the square brackets.
        [ -~]              All characters.
        [lO]               Lower case L and upper case o.

    Modifiers describe how the pattern matching works with the character
    set. There are six modifiers. Each one has a numeric value. They
    are listed below with their default values.

        Modifier      Default  Description
        ============  =======  ====================================
        min-matches         1  The min number of chars that match.
                               Zero means that if a positive match
                               occurs, the string is excluded. It
                               is used to create negative tests.

        max-matches         0  The max number of chars that match.
                               Zero means that there is no maximum.

        min-num-dups        0  The min number of chars that can be
                               duplicated.
                               Zero means that there is no minimum.

        max-num-dups        0  The max number of chars that can be
                               duplicated.
                               Zero means that there is no maximum.

        min-dup-size        2  The min number of chars in a
                               duplication.
                               Note that 1 doesn't make sense.
                               Anything less than 1 defaults to 2.

        max-dup-size        0  The max number of chars in a duplication
                               Zero means that there is no maximum.
    
    A character is duplicated if it occurs >= min-dup-size and <= max-dup-size
    times.

    Here is how the modifiers and the expressions work together to generate
    passwords with some reasonable characteristics.

       --length 8 \
       --charset '[a-z],min-matches=4' \
       --charset '[0-9],min-matches=1,max-matches=1' \
       --charset '[A-Z],min-matches=1' \
       --charset '[-&%,!@();.+_[]^~ ],min-matches=1,max-matches=1' \
       --charset '[Ol],min-matches=0'

    Here is how you might set things up to generate random user names.
    Notice how min-matches=0 was used to specify that the underscore
    should appear 0 or 1 times.

       --length 8 \
       --charset '[a-z],min-matches=7' \
       --charset '[_],min-matches=0,max-matches=1'

    """
    def __init__(self, arg):
        # Parse out the modifiers and the character set.
        b = arg.find('[')
        if b<0:
            raise Exception('character set not found, missing [ in %s' % (arg))
        if b>0 and arg[b-1] != ',':
            raise Exception('character set not found, missing , before [ in %s' % (arg))

        e = arg.rfind(']')
        if e<0:
            raise Exception('character set not found, missing ] in %s' % (arg))

        self._arg = arg
        self._expr = arg[b+1:e]
        prefix = arg[0:b]
        suffix = arg[e+1:]

        self._modstring = prefix+suffix

        # min-matches    the min number of chars that match
        # max-matches    the max number of chars that match
        # min-num-dups   the min number of chars that can be duplicated
        # max-num-dups   the max number of chars that can be duplicated
        # min-dup-size   the min number of chars in a duplication
        # max-dup-size   the max number of chars in a duplication
        #
        # A character is duplicated if it occurs 
        # >= min-dup-size and <= max-dup-size times.
        self._modifiers = {'min-matches':1, # set this to zero to exclude a set
                           'max-matches':0,
                           'min-num-dups':0,
                           'max-num-dups':0,
                           'min-dup-size':2,
                           'max-dup-size':0,
                           }
        self._charset = {}

    def Init(self):
        """
        Initialization after the object has been created.
        It can be chained because it returns self.
        """
        self._charset = {}
        self.InitParseExpr()
        self.InitParseModifiers()

        return self

    def Dump(self):
        """
        Dump information about the charset.
        """
        print
        print '  expr          : "%s"' % (str(self._expr))
        print '  num chars     : %d' % (len(self._charset))
        print '  num modifiers : %d' % (len(self._modifiers))
        print '  charset       : %s' % [(k, self._charset[k]) for k in sorted(self._charset.keys())]
        print '  modifiers     : %s' % [(k, self._modifiers[k]) for k in sorted(self._modifiers.keys())]
        print

    def GetCharsetString(self):
        """
        Get the charset as a string.
        """
        return ''.join(sorted(self._charset.keys()))

    def DupsAllowed(self):
        """
        This tells us whether sampling with replacement is required.
        """
        return self._modifiers['min-num-dups']>0 or self._modifiers['max-num-dups']>0

    def MeetsCriteria(self, string):
        """
        Tests the input string to see whether it meets the criteria
        for this charset.
        """
        num_matches = 0
        matched_char_cnt = {}
        for ch in string:
            if ch in self._charset:
                num_matches += 1
                # Count the number of times a specific character
                # occurs.
                if ch not in matched_char_cnt:
                    matched_char_cnt[ch] = 0
                matched_char_cnt[ch] += 1

        if self._modifiers['min-matches']>0:
            if num_matches < self._modifiers['min-matches']:
                return False
        if self._modifiers['max-matches']>0:
            if num_matches > self._modifiers['max-matches']:
                return False

        # Can't test individual character occurrences until all
        # of them have been counted.
        num_dups = 0 # number of duped chars
        for ch in matched_char_cnt:
            num = matched_char_cnt[ch] # number of dups of this char
            if num > 1:
                # Test individual character occurrences.
                if self._modifiers['min-dup-size']>0:
                    if num < self._modifiers['min-dup-size']:
                        return False
                if self._modifiers['max-dup-size']>0:
                    if num > self._modifiers['max-dup-size']:
                        return False

                # Update the dup count if it is in the range.
                num_dups += 1

        # Test all dups.
        if self._modifiers['min-num-dups']>0:
            if num_dups < self._modifiers['min-num-dups']:
                return False
        if self._modifiers['max-num-dups']>0:
            if num_dups > self._modifiers['max-num-dups']:
                return False

        return True

    def InitParseExpr(self):
        """
        Parse the charset expression.
        """
        expr = self._expr
        charset = {}

        # First grab the escaped '-'
        r = expr.find('--')
        while r>=0:
            if '-' not in charset:
                charset['-'] = 0
            charset['-'] += 1
            pre = expr[0:r]
            suf = expr[r+2:]
            expr = pre+suf
            r = expr.find('--')

        # Now capture the ranges: <ch>-<ch>.
        r = expr.find('-')
        while r>0:
            b = expr[r-1]
            e = expr[r+1]
            if ord(b)<=ord(e):
                beg = ord(b)
                end = ord(e)
            else:
                beg = ord(e)
                end = ord(b)
            for x in range(beg,end+1):
                ch = chr(x)
                if ch not in charset:
                    charset[ch] = 0
                charset[ch] += 1
            pre = expr[0:r-1]
            suf = expr[r+2:]
            expr = pre+suf
            r = expr.find('-')

        # Remaining characters.
        for ch in expr:
            if ch not in charset:
                charset[ch] = 0
                charset[ch] += 1

        self._charset = charset

    def InitParseModifiers(self):
        """
        Parse modifiers.
        """
        mods = self._modstring
        for mod in sorted(mods.split(',')):
            if mod not in ['',None]:
                # Ignore empty modifiers.
                # Allow two forms: ,a,b=c
                key = mod
                val = None
                if '=' in mod:
                    a = mod.split('=')
                    if len(a)!=2:
                        raise Exception('invalid modifier %s found in %s' % (mod,self._arg))
                    key = a[0]
                    val = a[1]
                else:
                    raise Exception('missing value for %s in %s' % (key,self._arg))
                
                key = key.lower()
                if key in self._modifiers:
                    m = re.search('^\d+$',val)
                    if not m:
                        raise Exception('numeric value for %s not found in %s' % (key,self._arg))
                    self._modifiers[key] = int(val)
                else:
                    raise Exception('unrecognized modifier %s found in %s' % (key,self._arg))

        if self._modifiers['min-dup-size'] < 2:
            self._modifiers['min-dup-size'] = 0

def GenerateAlphabet(charsets):
    """
    Generate the alphabet for the character sets.
    Eliminate the duplicate characters.
    This is used internally by KeyGenerator().

    Here is an example that uses the defaults:

       length = 8

       try:
           charsets = KeyGenerator.LoadDefaultCharsets(length)
       except Exception as (strerr):
           print "ERROR: charset compilation failed: %s" % (strerr)

       passwd = KeyGenerator.KeyGenerator(charsets,length)
       if not passwd:
           print "ERROR: password not generated!"
           sys.exit(1)

    @param charsets  The list of character sets.
    @returns The alphabet, a list of characters.
    """
    us = {}
    for charset in charsets:
        for ch in charset._charset.keys():
            us[ch] = 0
    alphabet = ''.join(sorted(us.keys()))
    return alphabet

def LoadDefaultCharsets(length=8):
    """
    Load the default character sets.

    Here is an example that uses the defaults:

       length = 8

       try:
           charsets = KeyGenerator.LoadDefaultCharsets(length)
       except Exception as (strerr):
           print "ERROR: charset compilation failed: %s" % (strerr)

       passwd = KeyGenerator.KeyGenerator(charsets,length)
       if not passwd:
           print "ERROR: password not generated!"
           sys.exit(1)

    @param length The length of the key/password.
    @returns The character sets.
    """
    # Specify the default character sets.
    charsets=[]

    half = max(1,length/2)
    charset = CharSet('[a-z],min-matches=%d' % (half)).Init()
    charsets.append(charset)

    charset = CharSet('[0-9],min-matches=1,max-matches=1').Init()
    charsets.append(charset)
    
    charset = CharSet('[A-Z],min-matches=1').Init()
    charsets.append(charset)
    
    charset = CharSet('[-&%,!@();.+_[]^~ ],min-matches=1,max-matches=1').Init()
    charsets.append(charset)
    
    charset = CharSet('[Ol],min-matches=0').Init()
    charsets.append(charset)

    return charsets

def SampleWithReplacement(alphabet,length):
    """
    Random.sample() samples without replacement which can cause
    problems for some character sets.
    @param alphabet  The alphabet to sample from.
    @param length    The size of the results.
    @returns The result.
    """
    n = len(alphabet)
    result = [None] * length
    for i in xrange(length):
        j = int(random.random() * n)
        result[i] = alphabet[j]
    return result

def KeyGenerator(charsets,length,maxtries=10000,debug=False):
    """
    Generate a password from that meets the criteria of all of the
    character sets.

    Here is an example using the default character sets.

       length = 8
       try:
           charsets = KeyGenerator.LoadDefaultCharsets(length)
       except Exception as (strerr):
           print "ERROR: charset compilation failed: %s" % (strerr)

       passwd = KeyGenerator.KeyGenerator(charsets,length)
       if not passwd:
           print "password not generated!"
           sys.exit(1)

    Here is an example that will generate random user names.

       # one character set of the lower case alphabetic characters.
       length = 8
       charsets = []
       try:
           charset.append(CharSet('[a-z],min-matches=%d' % (length)).Init())
       except Exception as (strerr):
           print "ERROR: charset compilation failed: %s" % (strerr)

       passwd = KeyGenerator.KeyGenerator(charsets,length)
       if not passwd:
           print "password not generated!"
           sys.exit(1)

    @param charsets The list of charsets.
    @param length   The length of the password.
    @param maxtries The maximum number of attempts.
    @returns The password if one was found or None.
    """
    alphabet = GenerateAlphabet(charsets)
    if debug:
        print "DEBUG: alphabet: "+alphabet

    replacement=False
    for charset in charsets:
        if charset.DupsAllowed():
            replacement=True
            break;

    if debug:
        print "DEBUG: sample with replacement: "+str(replacement)
    tries = 0
    done = False
    pw = None
    while not done:
        tries += 1
        if tries>maxtries:
            done = True

        # Generate a password using the alphabet.
        # If duplicates are allowed then we need to
        # sample with replacement.
        if replacement:
            pw=''.join(SampleWithReplacement(alphabet,length))
        else:
            pw=''.join(random.SystemRandom(random).sample(alphabet,length))
        if debug:
            print "DEBUG: testing: "+pw
        
        # Test whether it meets all of the criteria.
        done = True
        for charset in charsets:
            if not charset.MeetsCriteria(pw):
                done = False
                break

    return pw
    
