# ================================================================
# Name:      KeyStrength.py
# Purpose:   Key strength analyzer.
# Author:    Joe Linoff
# Created:   2012-01-10
# Id:        $Id: KeyStrength.py,v 1.5 2012/01/13 02:54:18 jlinoff Exp jlinoff $
# Copyright: (c) 2012 Joe Linoff
# License:
#     SecEd is free software; you can redistribute 
#     it and/or modify it under the terms of the GNU General Public License as 
#     published by the Free Software Foundation; either version 2 of the License, 
#     or (at your option) any later version.
#            
#     SecEd is distributed in the hope that it will be useful, 
#     but WITHOUT ANY WARRANTY; without even the implied warranty of 
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#     See the GNU General Public License for more details. You should have 
#     received a copy of the GNU General Public License along with SecEd; 
#     if not, write to the Free Software Foundation, Inc., 59 Temple Place, 
#     Suite 330, Boston, MA  02111-1307  USA
# ================================================================

import math

class KeyStrength:
    """
    Analyze key strength. It is used as follows:

       ks = KeyStrength(key,aps)
       keylen = len(key)
       cat, base, comb, maxttb = ks.GetCombinations()
       yrs, days, hrs, mins, secs = ks.GetTimeInfo(maxttb)
       rating, desc = ks.GetRating(cat, maxttb)
       print rating
       print desc
       print '%-16s : %s years, %s days, %02.0f:%02.0f:%02.0f\n' % ('MaxTimeFmt',
                                                                     ks.Commaize(yrs),
                                                                     days,hrs,mins,secs)
    """
    def __init__(self, key, aps):
        self._key = key # key value
        self._aps = aps # attempts per second
        self._secs_per_minute     = 60
        self._secs_per_hour    = self._secs_per_minute*60
        self._secs_per_day     = self._secs_per_hour*24
        self._secs_per_week    = self._secs_per_day*7
        self._secs_per_month   = self._secs_per_day*30
        self._secs_per_year    = self._secs_per_day*365.25
        self._secs_per_decade  = self._secs_per_year*10
        self._secs_per_century = self._secs_per_year*100

    def AnalyzeAndReport(self):
        key = self._key
        aps = self._aps
        keylen = len(key)
        cat, base, comb, maxttb = self.GetCombinations()
        yrs, days, hrs, mins, secs = self.GetTimeInfo(maxttb)
        rating, desc = self.GetRating(cat, maxttb)

        data =  '%-16s : %s\n' % ('Len',self.Commaize(keylen))
        data += '%-16s : %d\n' % ('Base',base)
        data += '%-16s : %d\n' % ('  Digits'     ,cat['charDigits'][0])
        data += '%-16s : %d\n' % ('  AlphaLC'    ,cat['charAlphaLC'][0])
        data += '%-16s : %d\n' % ('  AlphaUC'    ,cat['charAlphaUC'][0])
        data += '%-16s : %d\n' % ('  WhiteSpace' ,cat['charWS'][0])
        data += '%-16s : %d\n' % ('  Other'      ,cat['charOther'][0])
        data += '%-16s : %s\n' % ('Combinations'  ,self.Commaize(comb))
        data += '%-16s : %s\n' % ('AttemptsPerSec',self.Commaize(aps))
        data += '%-16s : %s seconds\n' % ('MaxTime',self.Commaize(maxttb))
        data += '%-16s : %s years, %s days, %02.0f:%02.0f:%02.0f\n' % ('MaxTimeFmt',
                                                                       self.Commaize(yrs),
                                                                       days,hrs,mins,secs)
        data += '%-16s : %s\n' % ('Rating',rating)
        data += '%-16s : %s\n' % ('RatingDesc',desc)
        return data

    def GetCombinations(self):
        key = self._key
        aps = self._aps

        cat = {'charDigits' :[0,10],
               'charAlphaLC':[0,26],
               'charAlphaUC':[0,26],
               'charWS'     :[0,1],
               'charOther'  :[0,32],}
        for ch in key:
            if ord(ch)>=ord('0') and ord(ch)<=ord('9'):
                cat['charDigits'][0] += 1
            elif ord(ch)>=ord('a') and ord(ch)<=ord('z'):
                cat['charAlphaLC'][0] += 1
            elif ord(ch)>=ord('A') and ord(ch)<=ord('Z'):
                cat['charAlphaUC'][0] += 1
            elif ch in [' ']:
                cat['charWS'][0] += 1
            elif ord(ch)>ord(' ') and ord(ch)<127:
                cat['charOther'][0] += 1

        base = 0
        for k in cat:
            if cat[k][0]:
                base += cat[k][1]

        comb = 1
        for i in range(len(key)):
            comb *= base

        maxttb = comb / aps

        return cat, base, comb, maxttb

    def GetTimeInfo(self, maxttb):
        remaining_secs = maxttb
        yrs  = math.floor(remaining_secs/self._secs_per_year)
        remaining_secs -= math.ceil(yrs*self._secs_per_year)
        days = math.floor(remaining_secs/self._secs_per_day)
        remaining_secs -= math.ceil(days*self._secs_per_day)
        hrs = math.floor(remaining_secs/self._secs_per_hour)
        remaining_secs -= math.ceil(hrs*self._secs_per_hour)
        mins = math.floor(remaining_secs/self._secs_per_minute)
        remaining_secs -= math.ceil(mins*self._secs_per_minute)
        secs = math.floor(remaining_secs)

        return yrs, days, hrs, mins, secs

    def GetRating(self, cat, maxttb):
        key = self._key
        aps = self._aps

        # Categorize.
        catAllAlphaLC = False
        catAllAlphaLC_WS = False
        catAllAlphaUC = False
        catAllAlphaUC_WS = False
        if cat['charDigits'][0] == 0 and cat['charOther'][0] == 0:
            if cat['charAlphaLC'][0] == 0 and cat['charAlphaUC'][0] != 0:
                if cat['charWS'][0] == 0:
                    catAllAlphaUC = True
                else:
                    catAllAlphaUC_WS = True
            elif cat['charAlphaLC'][0] != 0 and cat['charAlphaUC'][0] == 0:
                if cat['charWS'][0] == 0:
                    catAllAlphaLC = True
                else:
                    catAllAlphaLC_WS = True
            
        # Threshhold for max word size for dictionary attack.
        TH=10

        # I need to think about attacks for keys that only consist of
        # WS and words consisting of alphabetic characters because it
        # is highly probable that the words are subject to a
        # dictionary attack. For now I will combine the probabilites
        # of each word.
        # These are the factors: catAllAlphaLC_WS or catAllAlphaUC_WS:

        if catAllAlphaLC and len(key)<TH:
            return 'Very Weak',  'lower case dictionary attack'
        elif catAllAlphaUC and len(key)<TH:
            return 'Very Weak', 'upper case dictionary attack'
        elif maxttb>=self._secs_per_decade*1000:
            return 'Super Strong', 'more than a millenium'
        elif maxttb>=self._secs_per_decade*100:
            return 'Super Strong', 'more than a century'
        elif maxttb>=self._secs_per_decade:
            return 'Super Strong', 'more than a decade'
        elif maxttb>=self._secs_per_year:
            return 'Very Strong', 'more than a year but less than a decade'
        elif maxttb>=self._secs_per_month:
            return 'Strong', 'more than a month but less than a year'
        elif maxttb>=self._secs_per_week:
            return 'Moderate', 'more than a week but less than a month'
        elif maxttb>=self._secs_per_day:
            return 'Weak', 'more than a day but less than a week'
        elif maxttb>=self._secs_per_hour:
            return 'Very Weak', 'more than a hour but less than a day'
        elif maxttb>=self._secs_per_minute:
            return 'Super Weak', 'more than a minute but less than an hour'
        else:
            return 'Super Weak', 'less than a minute'

    def Commaize(self, n):
        """
        Insert commas into numbers.
        """
        s = '%.0f' % (n)
        l = len(s)
        if l>3:
            r = ''
            i = 0
            j = (l%3)
            for c in s:
                if i>0 and (i%3)==j:
                    r += ','
                r += c
                i += 1
            return r
        return s


