#!/usr/bin/env python2.7
# ================================================================
# Name:      SecEd.py
# Purpose:   Text editor for SecEd. Based on Peter Yared's 
#            pydocview example (PyDocView).
# Author:    Joe Linoff
# Created:   2012-01-10
# Id:        $Id: SecEd.py,v 1.8 2012/02/04 02:05:01 jlinoff Exp jlinoff $
# Copyright: (c) 2012 Joe Linoff
# License:
#     SecEd is free software; you can redistribute 
#     it and/or modify it under the terms of the GNU General Public License as 
#     published by the Free Software Foundation; either version 2 of the License, 
#     or (at your option) any later version.
#            
#     SecEd is distributed in the hope that it will be useful, 
#     but WITHOUT ANY WARRANTY; without even the implied warranty of 
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#     See the GNU General Public License for more details. You should have 
#     received a copy of the GNU General Public License along with SecEd; 
#     if not, write to the Free Software Foundation, Inc., 59 Temple Place, 
#     Suite 330, Boston, MA  02111-1307  USA
# ================================================================

import struct
import sys
import os.path
import time
import re
import wx
import wx.lib.docview as docview
import wx.lib.pydocview as pydocview
import TextEditor
import FindService
from KeyStrengthDialog import KeyStrengthDialog
from KeyGeneratorDialog import KeyGeneratorDialog
from DisplayStaticTextDialog import DisplayStaticTextDialog
_ = wx.GetTranslation

# ================================================================
# Classes
# ================================================================

class SecEd(pydocview.DocApp):
    """
    Top level object.
    """

    SPLASH = sys.path[0] + os.sep + "splash.png"
    HELP=sys.path[0] + os.sep + 'help.txt'
    TIPS=sys.path[0] + os.sep + 'tips.txt'
    LICENSE=sys.path[0] + os.sep + 'LICENSE.txt'
    SPLASH_MIN_TIME = 1
    APP_NAME = "SecEd"
    
    def OnInit(self):
        # Call the super - this is important!!!
        pydocview.DocApp.OnInit(self)

        # Show the splash dialog while everything is loading up
        if os.path.exists(SecEd.SPLASH):
            self.ShowSplash(SecEd.SPLASH)
            splash_time = time.clock()

        # Set the name and the icon
        self.SetAppName(_(SecEd.APP_NAME))
        self.SetDefaultIcon(pydocview.getBlankIcon())
        
        # Initialize the document manager
        docManager = docview.DocManager(flags = self.GetDefaultDocManagerFlags())  
        self.SetDocumentManager(docManager)

        # Create a template for text documents and associate it with the docmanager
        textTemplate = docview.DocTemplate(docManager,
                                           _("Text"),
                                           "*.text;*.txt",
                                           _("Text"),
                                           _(".txt"),
                                           _("Text Document"),
                                           _("Text View"),
                                           TextEditor.TextDocument,
                                           TextEditor.TextView,
                                           icon=pydocview.getBlankIcon())
        docManager.AssociateTemplate(textTemplate)

        # Install services - these can install menu and toolbar items
        textService           = self.InstallService(TextEditor.TextService())
        findService           = self.InstallService(FindService.FindService())
        optionsService        = self.InstallService(pydocview.DocOptionsService(supportedModes=wx.lib.docview.DOC_MDI))
        windowMenuService     = self.InstallService(pydocview.WindowMenuService())
        filePropertiesService = self.InstallService(pydocview.FilePropertiesService())
            
        # Install the TextEditor's option panel into the OptionsService
        optionsService.AddOptionsPanel(TextEditor.TextOptionsPanel)
        
        # Display the splash for the minimum amount of time.
        if os.path.exists(SecEd.SPLASH):
            splash_time = time.clock() - splash_time
            if splash_time<SecEd.SPLASH_MIN_TIME:
                x = SecEd.SPLASH_MIN_TIME - splash_time
                time.sleep(x)

        # If it is an MDI app open the main frame
        self.OpenMainFrame()
        
        # Open any files that were passed via the command line
        self.OpenCommandLineArgs()
        
        # If nothing was opened and it is an SDI app, open up an empty text document
        if not docManager.GetDocuments() and docManager.GetFlags() & wx.lib.docview.DOC_SDI:
            textTemplate.CreateDocument('', docview.DOC_NEW).OnNewDocument()

        # Install custom stuff.
        frame = self.GetTopWindow()
        menuBar = frame.GetMenuBar()
        helpMenu = menuBar.GetMenu(menuBar.FindMenu(_("&Help")))
        toolMenu = menuBar.GetMenu(menuBar.FindMenu(_("&Tools")))

        if helpMenu:
            # Get rid of the default About dialog, it is lame.
            item = helpMenu.FindItemByPosition(0)
            helpMenu.DestroyItem(item)
            
        if not helpMenu:
            helpMenu = wx.Menu()
            menuBar.Append(helpMenu.GetId(), _('&Help'))

        # Add my own custom About dialog box based on the
        # new wx.AboutDialog. The old one was clumsy.
        i = wx.NewId()
        helpMenu.Append(i,_('&Info'))
        wx.EVT_MENU(self,i,self.OnHelpInfo)

        helpMenu.AppendSeparator()

        i = wx.NewId()
        helpMenu.Append(i,_('&About'))
        wx.EVT_MENU(self,i,self.OnAboutBox)

        # Insert the new menu items at the front.
        toolMenu.InsertSeparator(0)
        i = wx.NewId()
        toolMenu.Insert(0,i,_('Key &Generator...\tCtrl+G'))
        wx.EVT_MENU(self,i,self.OnKeyGenerator)

        i = wx.NewId()
        toolMenu.Insert(0,i,_('&Key Strength...\tCtrl+K'))
        wx.EVT_MENU(self,i,self.OnKeyStrength)

        # Close the splash dialog
        if os.path.exists(SecEd.SPLASH):
            self.CloseSplash()
        
        # Show the tips dialog
        if os.path.exists(SecEd.TIPS):
            wx.CallAfter(self.ShowTip,
                         wx.GetApp().GetTopWindow(),
                         wx.CreateFileTipProvider(SecEd.TIPS, 0))

        # Overhead of updating menus was too much.
        # Change to update every N milliseconds.
        wx.UpdateUIEvent.SetUpdateInterval(1000)

        #
        #self.MyPopup(sys.platform+' '+str(sys.platform.startswith('win')),'PLATFORM')
 
        # Tell the framework that everything is great
        return True

    def OnKeyStrength(self,e):
        """
        Key Strength dialog.
        """
        dlg = KeyStrengthDialog(self.GetTopWindow(), -1, 'Key Strength Dialog', size=(700,450))
        dlg.Center()
        dlg.Show()

    def OnKeyGenerator(self,e):
        """
        Key Generator dialog.
        """
        dlg = KeyGeneratorDialog(self.GetTopWindow(), -1, 'Key Generator Dialog', size=(700,450))
        dlg.Center()
        dlg.Show()

    def OnHelpInfo(self,e):
        """
        Information.
        """
        if os.path.exists(SecEd.HELP):
            fp = open(SecEd.HELP,'r')
            data = fp.read()
            fp.close()
        else:
            data = 'Help file not found.'
        dialog = DisplayStaticTextDialog(self.GetTopWindow(), 'SecEd Help')
        dialog.Center()
        dialog.SetValue(data)
        dialog.ShowModal()
        dialog.Destroy()

    def OnAboutBox(self, e):
        """
        Use the new About box as described here: http://zetcode.com/wxpython/dialogs/
        """
        description = """
Secure Editor (SecEd) allows you to read or modify the contents of
secure files as text. It also allows you to score security of your
encryption key. You can use it to store password and other sensitive
information.

SecEd also allows you test the strength of key/password using
the 'Tools -> Key Strength...' dialog.

It was written entirely in Python 2.7 using the wxPython library
for the GUI display.

%s %d
""" % (sys.platform, struct.calcsize("P")*8)

        if os.path.exists(SecEd.LICENSE):
            fp = open(SecEd.LICENSE,'r')
            license = fp.read()
            fp.close()
        else:
            license = ''

        info = wx.AboutDialogInfo()
        SPLASH = SecEd.SPLASH # Use the class variable value
        if os.path.exists(SPLASH):
            info.SetIcon(wx.Icon(SPLASH, wx.BITMAP_TYPE_PNG))
        info.SetName('SecEd')
        info.SetVersion('0.9.1')
        info.SetDescription(description)
        info.SetCopyright('(C) 2012 - Joe Linoff')
        if license != '':
            info.SetLicense(license)
        info.AddDeveloper('Joe Linoff')
        info.AddDocWriter('Joe Linoff')
        wx.AboutBox(info)

    def MyPopup(self, msg, title):
        """
        Pops up a dialogue with a message.
        Used for internal debugging.
        """
        topWin = self.GetTopWindow()
        dlg = wx.MessageDialog(topWin,msg,title,wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def DebugPopup(self, msg):
        """
        Pops up a dialogue with a message.
        Used for internal debugging.
        """
        self.MyPopup(msg,'Debug Message')

    def ErrorPopup(self, msg, _exit=True):
        """
        Pops up a dialogue with a message.
        Used for displaying errors.
        """
        self.MyPopup('ERROR: '+msg,'Error Message')
        if _exit:
            sys.exit(1)

#----------------------------------------------------------------------------
# Main
#----------------------------------------------------------------------------

# Run the SecEd and do not redirect output to the wxPython error dialog
app = SecEd(redirect=False)
app.MainLoop()
