#!/usr/bin/env python2.7
import getpass
import getopt
import os
import sys
import re
import datetime
import shlex, subprocess
import bz2
import base64
import tempfile
import time
import math
import KeyStrength

def usage():
    """
    Print the program help and exit.
    """
    fn = os.path.basename(__file__)
    print """
USAGE
    %s [OPTIONS]

DESCRIPTION
    Report key strength.

    This tool looks at the key contents and reports how strong it is.

    The ratings are described in the table below for a rainbow attack
    in increasing order of strength.

       Rating        Description
       ============  ======================================
       Super Weak    Trivial to break (less than an hour).

       Very Weak     Less than a day, more than an hour.

       Weak          Subject to a dictionary attack or
                     a rainbow attack taking less than a week
                     but more than a day.

       Moderate      More than week, less than a month.

       Strong        More than a month, less than 6 months.

       Very Strong   More than 6 months, less than a year.

       Super Strong  More than 1 year.
       
    They are completely dependent on the number of attempts per
    second that an attacker can carry out. 

    The default is 10,000 based on the assumption that it takes 1ms to
    perform a computation and comparison on a modern 3GHz processor
    with 10 or more threads.

    Furthermore, if we assume the use of 10 parallel processors, it is
    fairly simple to get to 100,000 operations per second. Dedicated
    HW based on an FPGA would be even faster.

    Of course a dedicated cracking system could use thousands of such
    processors or dedicated HW to achieve rates of more than 1,000,000
    per second.

    You can adjust this value by specifying the -a argument.

OPTIONS
    There are short and long forms for most options.

    -a <num>, --attempts-per-second
                 The maximimum number of crack attempts per
                 second. If not specified the default is
                 10000.

    -k <key>, --key <key>
                 The key to analyze.

    -h, --help   Display this usage message.

    -v, --verbose
                 The level of verbosity. It can be specified
                 multiple times.

    -V, --version
                 Print the program version and exit.

EXAMPLES
    %% %s -h

    %% # Analyze a key
    %% %s
    Enter key:
    <output snipped>

    %% # Analyze a key, assume 1,000,000 crack attempts per second
    %% %s -a 1000000
    Enter key:
    <output snipped>

    %% # Analyze a key from the CLI
    %% %s -k '1aB!'
    Length           : 4
    Base             : 94
      Digits         : 1
      AlphaLc        : 1
      AlphaUc        : 1
      WhiteSpace     : 0
      Other          : 1
    Combinations     : 78,074,896
    AttemptsPerSec   : 10,000
    MaxTime          : 7,807 seconds
    MaxTimeFmt       : 0 years, 0 days, 02:10:07
    Rating           : Very Weak (less than a day, more than an hour)

AUTHOR
    Joe Linoff (jlinoff@tabula.com)

VERSION
    $Id: seced-key-strength.py,v 1.7 2012/01/11 18:49:22 jlinoff Exp $

COPYRIGHT
    %s allows you to decrypt files.

    Copyright (c) 2011 by Joe Linoff

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    """ % (fn,fn,fn,fn,fn,fn)
    sys.exit(0)

def commaize(n):
    """
    Insert commas into numbers.
    """
    s = '%.0f' % (n)
    l = len(s)
    if l>3:
        r = ''
        i = 0
        j = (l%3)
        for c in s:
            if i>0 and (i%3)==j:
                r += ','
            r += c
            i += 1
        return r
    return s

def main():
    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:],
                                       "a:hk:vV",
                                       ["attempts-per-second=",
                                        "help",
                                        "key=",
                                        "verbose",
                                        "version"])
    except getopt.GetoptError, err:
        print str(err)
        print "exiting ..."
        sys.exit(1)

    verbose=0
    key=None
    aps = 10000
    for opt,arg in opts:
        if opt in ["-h","--help"] :
            usage()
        elif opt in ["-k","--key"] :
            key=arg
        elif opt in ["-a","--attempts-per-second"] :
            aps=float(arg)
        elif opt in ["-v","--verbose"] :
            verbose += 1
        elif opt in ["-V","--version"] :
            print 'Version: $Id: seced-key-strength.py,v 1.7 2012/01/11 18:49:22 jlinoff Exp $'
            sys.exit(0)
        else:
            print "ERROR: Unrecognized option "+opt
            sys.exit(1)
    
    fns = []
    for fn in args:
        fns.append(fn)

    if len(fns) > 0:
        print "ERROR: unexpected arguments."
        sys.exit(1)

    if not key:
        key=getpass.getpass('Enter key: ')

    keylen = len(key)
    ks = KeyStrength.KeyStrength(key,aps)
    print ks.AnalyzeAndReport()

if __name__ == "__main__":
    main()
