#!/usr/bin/env python2.7
import getopt
import sys
import re
import os

import KeyGenerator

def usage():
    """Print out the script usage."""
    fn = os.path.basename(__file__)
    print """
USAGE
    %s [OPTIONS]

DESCRIPTION
    Generate one or more strong random passwords of fixed length.
    
    By default the character does not contain lower case L (l) or
    upper case o () so no passwords that contain lower case L (l) or
    upper case o (O) are generated because lower case L is hard to
    differentiate from 1 and upper case o is hard to differentiate
    from zero (0).

    You can override this by defining your own character set.

    Character sets are the key to making this. They are specified by a
    comma separated tuple of modifiers and a single expression in a
    string. Here is an example that will explained later.

        # accept digits, lower case letters, upper case letters,
        # dollar sign and period. At least 3 must appear.
        charset #1: '[0-9a-bA-Z$.]',min-matches=3

        # exclude lower case L and upper case o
        charset #2: '[lO]',min-matches=0

    The expression consists of printable characters in the ASCII range
    of 32 (' ') to 126 ('~') enclosed in brackets. A range can be
    specified using a dash ('-'). A dash is specified by inserting it
    at the front. Internal brackets do not have to be escaped.

    Here are a couple examples of expressions:

        Expression     Description
        =============  ======================================
        [0123456789]   All digits.
        [0-9]          All digits.
        [a-z0-9]       All lower case letters and all digits.
        [-[]]          A dash and the square brackets.
        [ -~]          All characters.
        [lO]           Lower case L and upper case o.

    Modifiers describe how the pattern matching works with the character
    set. There are six modifiers. Each one has a numeric value. They
    are listed below with their default values.

        Modifier      Default  Description
        ============  =======  ====================================
        min-matches         1  The min number of chars that match.
                               Zero means that if a positive match
                               occurs, the string is excluded. It
                               is used to create negative tests.

        max-matches         0  The max number of chars that match.
                               Zero means that there is no maximum.

        min-num-dups        0  The min number of chars that can be
                               duplicated.
                               Zero means that there is no minimum.

        max-num-dups        0  The max number of chars that can be
                               duplicated.
                               Zero means that there is no maximum.

        min-dup-size        2  The min number of chars in a
                               duplication.
                               Note that 1 doesn't make sense.
                               Anything less than 1 defaults to 2.

        max-dup-size        0  The max number of chars in a duplication
                               Zero means that there is no maximum.
    
    A character is duplicated if it occurs >= min-dup-size and
    <= max-dup-size times.

    Here is how the modifiers and the expressions work together to
    generate passwords with some reasonable characteristics.

       --length 8 \\
       --charset '[a-z],min-matches=4' \\
       --charset '[0-9],min-matches=1,max-matches=1' \\
       --charset '[A-Z],min-matches=1' \\
       --charset '[-&%%,!@();.+_[]^~ ],min-matches=1,max-matches=1' \\
       --charset '[Ol],min-matches=0'

    Here is how you might set things up to generate random user names.
    Notice how min-matches=0 was used to specify that the underscore
    should appear 0 or 1 times.

       --length 8 \\
       --charset '[a-z],min-matches=7' \\
       --charset '[_],min-matches=0,max-matches=1'

OPTIONS
    -c <set>     Specify a character set to use.
                 It can be specified multiple times.
                 If it is not specified, the following
                 setup is used as the default to generate
                 reasonable passwords:

                     --length 8 \\
                     --charset '[a-z],min-matches=4' \\
                     --charset '[0-9],min-matches=1,max-matches=1' \\
                     --charset '[A-Z],min-matches=1' \\
                     --charset '[-&%%,!@();.+_[]^~ ],min-matches=1,max-matches=1' \\
                     --charset '[Ol],min-matches=0'

    -h           Display this usage message.

    -l <length>  The length of the password.
                 The default is 8.

    -n <num>     The number of passwords to generate.
                 The default is 1.

    -v           Increase the level of verbosity.
                 This can be specified multiple times.

    -V           Print the program version and exit.

EXAMPLE
    %% %s
    3&MEUmi0

    %% # Generate 5 passwords using the defaults.
    %% %s -n 5
    @05sNcqW
    j3Hix)1k
    bU;fc6kI
    b94df~AG
    t+IkVp29

    %% # Generate 5 user names each with length 8.
    %% # Notice that underscore only appears occasionally.
    %% %s --length 8 \\
          --charset '[a-z],min-matches=7' \\
          --charset '[_],min-matches=0,max-matches=1'
    msxlh_bc
    xmrcsdkl
    tlj_wgfo
    cetghkzn
    qvuiltbn

AUTHOR
    Joe Linoff (jlinoff@tabula.com)

VERSION
    $Id: seced-keygen.py,v 1.1 2012/02/04 01:58:29 jlinoff Exp $

COPYRIGHT
    %s allows you to encrypt files.

    Copyright (c) 2011 by Joe Linoff

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
""" % (fn,fn,fn,fn,fn)
    sys.exit(0)

def err(msg):
    """
    Print an error message and exit.
    """
    print 'ERROR: '+str(msg)
    sys.exit(1)

def warn(msg):
    """
    Print a warning message.
    """
    print 'WARNING: '+str(msg)

def info2(key,data):
    """
    Print info.
    """
    print '%-16s : %s' % (key,str(data))

def main():
    """main"""
    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:],
                                       "c:hl:m:n:vV",
                                       ["charset=",
                                        "help",
                                        "length=",
                                        "maxtries=",
                                        "number=",
                                        "verbose",
                                        "version"])
    except getopt.GetoptError, errmsg:
        print str(errmsg)
        print "exiting ..."
        sys.exit(1)
        
    verbose=0
    maxtries=10000
    charsets=[]
    number=1
    length=8
    for opt,arg in opts:
        if opt in ["-h","--help"] :
            usage()
        elif opt in ["-c","--charset"] :
            try:
                charset = KeyGenerator.CharSet(arg).Init()
                charsets.append(charset)
            except Exception as (strerr):
                err('charset compilation failed: %s' % (strerr))
        elif opt in ["-l","--length"] :
            m = re.search('^\d+$',arg)
            if not m:
                err('invalid argument for '+opt+', expected a number')
            length = int(arg)
        elif opt in ["-m","--maxtries"] :
            m = re.search('^\d+$',arg)
            if not m:
                err('invalid argument for '+opt+', expected a number')
            maxtries = int(arg)
        elif opt in ["-n","--number"] :
            m = re.search('^\d+$',arg)
            if not m:
                err('invalid argument for '+opt+', expected a number')
            number = int(arg)
        elif opt in ["-v","--verbose"] :
            verbose += 1
        elif opt in ["-V","--version"] :
            print 'Version: $Id: seced-keygen.py,v 1.1 2012/02/04 01:58:29 jlinoff Exp $'
            sys.exit(0)
        else:
            err("Unrecognized option "+opt)

    if len(charsets) == 0:
        try:
            charsets = KeyGenerator.LoadDefaultCharsets(length)
        except Exception as (strerr):
            err('ERROR: charset compilation failed: %s' % (strerr))

    if verbose:
        alphabet = KeyGenerator.GenerateAlphabet(charsets)
        info2('Alphabet',alphabet)
        info2('Length',length)
        info2('NumCharsets',len(charsets))

        if verbose>1:
            i = 0
            for cs in charsets:
                i+=1
                info2('Charset[%2d]'%i,'')
                cs.Dump()
        print

    for i in range(number):
        pw = KeyGenerator.KeyGenerator(charsets, length)
        print pw

if __name__ == "__main__":
    main()
